#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.commands`` -- Commands
"""

import sys

from edbob import commands

import rattail


class Command(commands.Command):
    """
    The primary command for Rattail.
    """
    
    name = 'rattail'
    version = rattail.__version__
    description = "Retail Software Framework"
    long_description = """
Rattail is a retail software framework.

Copyright (c) 2010-2012 Lance Edgar <lance@edbob.org>

This program comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.
See the file COPYING.txt for more information.
"""


class FileMonitorCommand(commands.FileMonitorCommand):
    """
    Interacts with the file monitor service; called as ``rattail filemon``.

    See :class:`edbob.commands.FileMonitorCommand` for more information.
    """

    def get_win32_module(self):
        from rattail import filemon
        return filemon

    def get_win32_service(self):
        from rattail.filemon import RattailFileMonitor
        return RattailFileMonitor


class InitCommand(commands.Subcommand):
    """
    Initializes the database; called as ``{{package}} initialize``.  This is
    meant to be leveraged as part of setting up the application.  The database
    used by this command will be determined by config, for example::

    .. highlight:: ini

       [edbob.db]
       sqlalchemy.url = postgresql://user:pass@localhost/{{package}}
    """

    name = 'initialize'
    description = "Initialize the database"

    def run(self, args):
        from edbob.db import engine
        from edbob.db.util import install_core_schema
        from edbob.db.exceptions import CoreSchemaAlreadyInstalled
        from edbob.db.extensions import activate_extension

        # Install core schema to database.
        try:
            install_core_schema(engine)
        except CoreSchemaAlreadyInstalled, err:
            print '%s:' % err
            print '  %s' % engine.url
            return

        # Activate any extensions you like here...
        # activate_extension('shrubbery')

        # Okay, on to bootstrapping...

        from edbob.db import Session
        from edbob.db.classes import Role, User
        from edbob.db.auth import administrator_role

        session = Session()

        # Create 'admin' user with full rights.
        admin = User(username='admin', password='admin')
        admin.roles.append(administrator_role(session))
        session.add(admin)

        # Do any other bootstrapping you like here...
        
        session.commit()
        session.close()
        
        print "Initialized database:"
        print '  %s' % engine.url


def main(*args):
    """
    The primary entry point for the Rattail command system.
    """

    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.extension.model`` -- Schema Definition
"""

from __future__ import absolute_import

import re

from sqlalchemy import (Column, String, Integer, Date, DateTime,
                        Boolean, Text, ForeignKey, BigInteger, Numeric)
from sqlalchemy import and_
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

import edbob
from edbob.db.model import Base, uuid_column
from edbob.db.extensions.contact import (Person, EmailAddress, PhoneNumber,
                                         PersonPhoneNumber)
from edbob.sqlalchemy import getset_factory


# __all__ = ['SilColumn', 'BatchDictionaryColumn', 'BatchDictionary',
#            'BatchTerminalColumn', 'BatchTerminal', 'BatchColumn',
#            'Batch', 'Brand', 'Department', 'Subdepartment', 'Category',
#            'Product', 'Employee', 'Vendor', 'VendorContact', 'VendorPhoneNumber',
#            'ProductCost', 'ProductPrice', 'Customer', 'CustomerGroup', 
#            'CustomerGroupAssignment', 'CustomerPerson']

__all__ = ['Department', 'Subdepartment', 'Brand', 'Category', 'Vendor',
           'VendorContact', 'VendorPhoneNumber', 'Product', 'ProductCost',
           'ProductPrice', 'Customer', 'CustomerEmailAddress',
           'CustomerPhoneNumber', 'CustomerGroup', 'CustomerGroupAssignment',
           'CustomerPerson', 'Employee', 'EmployeeEmailAddress',
           'EmployeePhoneNumber']

# sil_type_pattern = re.compile(r'^(CHAR|NUMBER)\((\d+(?:\,\d+)?)\)$')


# class SilColumn(Base):
#     """
#     Represents a SIL-compatible column available to the batch system.
#     """

#     __tablename__ = 'sil_columns'

#     uuid = uuid_column()
#     sil_name = Column(String(10))
#     display = Column(String(20))
#     data_type = Column(String(15))
    
#     def __repr__(self):
#         return "<SilColumn: %s>" % self.sil_name

#     def __str__(self):
#         return str(self.sil_name or '')


# class BatchDictionaryColumn(Base):
#     """
#     Represents a column within a :class:`BatchDictionary`.
#     """

#     __tablename__ = 'batch_dictionary_columns'

#     uuid = uuid_column()
#     dictionary_uuid = Column(String(32), ForeignKey('batch_dictionaries.uuid'))
#     sil_column_uuid = Column(String(32), ForeignKey('sil_columns.uuid'))
#     key = Column(Boolean)

#     sil_column = relationship(SilColumn)

#     def __repr__(self):
#         return "<BatchDictionaryColumn: %s>" % self.sil_column

#     def __str__(self):
#         return str(self.sil_column or '')


# class BatchDictionary(Base):
#     """
#     Represents a SIL-based dictionary supported by one or more
#     :class:`BatchTerminal` classes.
#     """

#     __tablename__ = 'batch_dictionaries'

#     uuid = uuid_column()
#     name = Column(String(20))
#     description = Column(String(255))

#     columns = relationship(
#         BatchDictionaryColumn,
#         backref='dictionary')

#     def __repr__(self):
#         return "<BatchDictionary: %s>" % self.name

#     def __str__(self):
#         return str(self.description or '')


# class BatchTerminalColumn(Base):
#     """
#     Represents a column supported by a :class:`BatchTerminal`.
#     """

#     __tablename__ = 'batch_terminal_columns'

#     uuid = uuid_column()
#     terminal_uuid = Column(String(32), ForeignKey('batch_terminals.uuid'))
#     dictionary_uuid = Column(String(32), ForeignKey('batch_dictionaries.uuid'))
#     sil_column_uuid = Column(String(32), ForeignKey('sil_columns.uuid'))
#     ordinal = Column(Integer)
#     source = Column(Boolean)
#     target = Column(Boolean)

#     dictionary = relationship(BatchDictionary)
#     sil_column = relationship(SilColumn)

#     def __repr__(self):
#         return "<BatchTerminalColumn: %s, %s, %s>" % (
#             self.terminal, self.dictionary, self.sil_column)

#     def __str__(self):
#         return str(self.sil_column or '')


# class BatchTerminal(Base):
#     """
#     Represents a terminal, or "junction" for batch data.
#     """

#     __tablename__ = 'batch_terminals'

#     uuid = uuid_column()
#     sil_id = Column(String(20), unique=True)
#     description = Column(String(50))
#     class_spec = Column(String(255))
#     functional = Column(Boolean, default=False)
#     source = Column(Boolean)
#     target = Column(Boolean)
#     source_kwargs = Column(Text)
#     target_kwargs = Column(Text)

#     columns = relationship(
#         BatchTerminalColumn,
#         backref='terminal')

#     _terminal = 'not_got_yet'

#     def __repr__(self):
#         return "<BatchTerminal: %s>" % self.sil_id

#     def __str__(self):
#         return str(self.description or '')

#     def source_columns(self, dictionary):
#         for col in self.columns:
#             if col.dictionary is dictionary:
#                 yield col

#     def get_terminal(self):
#         """
#         Returns the :class:`rattail.batches.BatchTerminal` instance which is
#         associated with the database record via its ``python_spec`` field.
#         """

#         if self._terminal == 'not_got_yet':
#             self._terminal = None
#             if self.class_spec:
#                 term = edbob.load_spec(self.class_spec)
#                 if term:
#                     self._terminal = term()
#         return self._terminal


# class BatchColumn(Base):
#     """
#     Represents a :class:`SilColumn` associated with a :class:`Batch`.
#     """

#     __tablename__ = 'batch_columns'

#     uuid = uuid_column()
#     batch_uuid = Column(String(32), ForeignKey('batches.uuid'))
#     ordinal = Column(Integer)
#     sil_column_uuid = Column(String(32), ForeignKey('sil_columns.uuid'))
#     source_uuid = Column(String(32), ForeignKey('batch_terminals.uuid'))
#     targeted = Column(Boolean)

#     sil_column = relationship(SilColumn)

#     source = relationship(
#         BatchTerminal,
#         primaryjoin=BatchTerminal.uuid == source_uuid,
#         order_by=[BatchTerminal.description],
#         )

#     def __repr__(self):
#         return "<BatchColumn: %s, %s>" % (self.batch, self.sil_column)


# def get_sil_column(name):
#     """
#     Returns a ``sqlalchemy.Column`` instance according to Rattail's notion of
#     what each SIL field ought to look like.
#     """

#     type_map = {

#         # The first list of columns is a subset of Level 1 SIL.

#         'F01':
#             Integer,    # upc
#         'F02':
#             String(60), # short (receipt) description
#         'F478':
#             Integer,    # scale text type

#         # The remaining columns are custom to Rattail.

#         'F4001':
#             String(60), # short description, line 2
#         }

#     return Column(name, type_map[name])


# def get_sil_type(data_type):
#     """
#     Returns a SQLAlchemy-based data type according to the SIL-compliant type
#     specifier found in ``data_type``.
#     """

#     if data_type == 'GPC(14)':
#         return BigInteger

#     m = sil_type_pattern.match(data_type)
#     if m:
#         data_type, precision = m.groups()
#         if precision.isdigit():
#             precision = int(precision)
#             scale = 0
#         else:
#             precision, scale = precision.split(',')
#             precision = int(precision)
#             scale = int(scale)
#         if data_type == 'CHAR':
#             assert not scale, "FIXME"
#             return String(precision)
#         if data_type == 'NUMBER':
#             return Numeric(precision, scale)

#     assert False, "FIXME"
    

# class Batch(Base):
#     """
#     Represents a batch of data, presumably in need of processing.
#     """

#     __tablename__ = 'batches'

#     _rowclass = None
#     _row_classes = {}

#     uuid = uuid_column()
#     source_uuid = Column(String(32), ForeignKey('batch_terminals.uuid'))
#     source_description = Column(String(50))
#     source_batch_id = Column(String(8))
#     batch_id = Column(String(8))
#     dictionary_uuid = Column(String(32), ForeignKey('batch_dictionaries.uuid'))
#     name = Column(String(30))
#     target_uuid = Column(String(32), ForeignKey('batch_terminals.uuid'))
#     action_type = Column(String(6))
#     elements = Column(String(255))
#     description = Column(String(50))
#     rowcount = Column(Integer, default=0)
#     effective = Column(DateTime)
#     deleted = Column(Boolean, default=False)
#     sil_type = Column(String(2))
#     sil_source_id = Column(String(20))
#     sil_target_id = Column(String(20))
#     sil_audit_file = Column(String(12))
#     sil_response_file = Column(String(12))
#     sil_origin_time = Column(DateTime)
#     sil_purge_date = Column(Date)
#     sil_user1 = Column(String(30))
#     sil_user2 = Column(String(30))
#     sil_user3 = Column(String(30))
#     sil_warning_level = Column(Integer)
#     sil_max_errors = Column(Integer)
#     sil_level = Column(String(7))
#     sil_software_revision = Column(String(4))
#     sil_primary_key = Column(String(50))
#     sil_sys_command = Column(String(512))
#     sil_dict_revision = Column(String(8))
    
#     source = relationship(
#         BatchTerminal,
#         primaryjoin=BatchTerminal.uuid == source_uuid,
#         order_by=[BatchTerminal.description],
#         )

#     dictionary = relationship(
#         BatchDictionary,
#         order_by=[BatchDictionary.name],
#         )

#     target = relationship(
#         BatchTerminal,
#         primaryjoin=BatchTerminal.uuid == target_uuid,
#         order_by=[BatchTerminal.description],
#         )

#     columns = relationship(
#         BatchColumn,
#         backref='batch',
#         )

#     # _table = None
#     # # _source_junction = 'not set'
#     # # _target_junction = 'not set'

#     # invalid_name_chars = re.compile(r'[^A-Za-z0-9]')

#     def __repr__(self):
#         return "<Batch: %s>" % (self.name or '(no name)')

#     def __str__(self):
#         return str(self.name or '')

#     # @property
#     # def source_junction(self):
#     #     """
#     #     Returns the :class:`rattail.BatchJunction` instance associated with
#     #     this batch's :attr:`Batch.source` attribute.
#     #     """

#     #     if self._source_junction == 'not set':
#     #         from rattail.sil import get_available_junctions
#     #         self._source_junction = None
#     #         junctions = get_available_junctions()
#     #         if self.source in junctions:
#     #             self._source_junction = junctions[self.source]
#     #     return self._source_junction

#     # @property
#     # def table(self):
#     #     """
#     #     Returns the ``sqlalchemy.Table`` instance for the underlying batch
#     #     data.
#     #     """

#     #     # from sqlalchemy import MetaData, Table, Column, String
#     #     from sqlalchemy import Table, Column, String
#     #     from rattail import metadata
#     #     from rattail.sqlalchemy import get_sil_column

#     #     if self._table is None:
#     #         # assert self.uuid
#     #         assert self.name
#     #         name = 'batch.%s.%s' % (self.source, self.batch_id)
#     #         if name in metadata.tables:
#     #             self._table = metadata.tables[name]
#     #         else:
#     #             # session = object_session(self)
#     #             # metadata = MetaData(session.bind)
#     #             columns = [Column('uuid', String(32), primary_key=True, default=get_uuid)]
#     #             # columns.extend([get_sil_column(x) for x in self.elements.split(',')])
#     #             columns.extend([get_sil_column(x) for x in self.elements.split(',')])
#     #             self._table = Table(name, metadata, *columns)
#     #     return self._table

#     # @property
#     # def rowclass(self):
#     #     """
#     #     Returns a unique subclass of :class:`rattail.BatchRow`, specific to the
#     #     batch.
#     #     """

#     #     if self._rowclass is None:
#     #         name = self.invalid_name_chars.sub('_', self.name)
#     #         self._rowclass = type('BatchRow_%s' % str(name), (BatchRow,), {})
#     #         mapper(self._rowclass, self.table)
#     #         # session = object_session(self)
#     #         # engine = session.bind
#     #         # session.configure(binds={self._rowclass:engine})
#     #     return self._rowclass

#     @property
#     def rowclass(self):
#         """
#         Returns the SQLAlchemy-mapped class for the underlying data table.
#         """

#         assert self.uuid
#         if self.uuid not in self._row_classes:
#             kwargs = {
#                 '__tablename__':        'batch.%s' % self.name,
#                 'uuid':                 uuid_column(),
#                 }
#             for col in self.columns:
#                 kwargs[col.sil_column.sil_name] = Column(get_sil_type(col.sil_column.data_type))
#             self._row_classes[self.uuid] = type('BatchRow', (Base,), kwargs)
#         return self._row_classes[self.uuid]

#     def create_table(self):
#         """
#         Creates the batch's data table within the database.
#         """

#         self.rowclass.__table__.create()

#     # @property
#     # def target_junction(self):
#     #     """
#     #     Returns the :class:`rattail.BatchJunction` instance associated with
#     #     this batch's :attr:`Batch.target` attribute.
#     #     """

#     #     if self._target_junction == 'not set':
#     #         from rattail.sil import get_available_junctions
#     #         self._target_junction = None
#     #         junctions = get_available_junctions()
#     #         if self.target in junctions:
#     #             self._target_junction = junctions[self.target]
#     #     return self._target_junction

#     # def append(self, **row):
#     #     """
#     #     Appends a row of data to the batch.  Note that this is done
#     #     immediately, and not within the context of any transaction.
#     #     """

#     #     # self.connection.execute(self.table.insert().values(**row))
#     #     # self.rowcount += 1
#     #     session = object_session(self)
#     #     session.add(self.rowclass(**row))
#     #     self.rowcount += 1
#     #     session.flush()

#     def add_rows(self, source, dictionary, **kwargs):
#         session = object_session(self)
#         source = source.get_terminal()
#         for row in source.provide_rows(session, self.rowclass,
#                                        dictionary, **kwargs):
#             session.add(row)
#             session.flush()

#     def execute(self):
#         """
#         Invokes the batch execution logic.  This will instantiate the
#         :class:`rattail.batches.BatchTerminal` instance identified by the
#         batch's :attr:`target` attribute and ask it to process the batch
#         according to its action type.

#         .. note::
#            No check is performed to verify the current time is appropriate as
#            far as the batch's effective date is concerned.  It is assumed that
#            other logic has already taken care of that and that yes, in fact it
#            *is* time for the batch to be executed.
#         """

#         target = self.target.get_terminal()
#         target.execute_batch(self)

#     def provide_rows(self):
#         """
#         Generator which yields :class:`BatchRow` instances belonging to the
#         batch.
#         """

#         session = object_session(self)
#         for row in session.query(self.rowclass):
#             yield row


# class BatchRow(edbob.Object):
#     """
#     Superclass of batch row objects.
#     """

#     def __repr__(self):
#         return "<BatchRow: %s>" % self.key_value


class Brand(Base):
    """
    Represents a brand or similar product line.
    """

    __tablename__ = 'brands'

    uuid = uuid_column()
    name = Column(String(100))

    def __repr__(self):
        return "<Brand: %s>" % self.name

    def __str__(self):
        return str(self.name or '')


class Department(Base):
    """
    Represents an organizational department.
    """

    __tablename__ = 'departments'

    uuid = uuid_column()
    number = Column(Integer)
    name = Column(String(30))

    def __repr__(self):
        return "<Department: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Subdepartment(Base):
    """
    Represents an organizational subdepartment.
    """

    __tablename__ = 'subdepartments'

    uuid = uuid_column()
    number = Column(Integer)
    name = Column(String(30))
    department_uuid = Column(String(32), ForeignKey('departments.uuid'))

    department = relationship(Department)

    def __repr__(self):
        return "<Subdepartment: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Category(Base):
    """
    Represents an organizational category for products.
    """

    __tablename__ = 'categories'

    uuid = uuid_column()
    number = Column(Integer)
    name = Column(String(50))
    department_uuid = Column(String(32), ForeignKey('departments.uuid'))

    department = relationship(Department)

    def __repr__(self):
        return "<Category: %s, %s>" % (self.number, self.name)

    def __str__(self):
        return str(self.name or '')

    def __unicode__(self):
        return unicode(self.name or '')


class VendorEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Vendor`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Vendor'}


class VendorPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Vendor`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Vendor'}


class VendorContact(Base):
    """
    Represents a point of contact (e.g. salesperson) for a vendor, from the
    retailer's perspective.
    """

    __tablename__ = 'vendor_contacts'

    uuid = uuid_column()
    vendor_uuid = Column(String(32), ForeignKey('vendors.uuid'), nullable=False)
    person_uuid = Column(String(32), ForeignKey('people.uuid'), nullable=False)
    preference = Column(Integer, nullable=False)

    person = relationship(Person)

    def __repr__(self):
        return "<VendorContact: %s, %s>" % (self.vendor, self.person)

    def __unicode__(self):
        return unicode(self.person)


class Vendor(Base):
    """
    Represents a vendor from which products are purchased by the store.
    """

    __tablename__ = 'vendors'

    uuid = uuid_column()
    id = Column(String(15))
    name = Column(String(40))

    contacts = relationship(VendorContact, backref='vendor',
                            collection_class=ordering_list('preference', count_from=1),
                            order_by=VendorContact.preference)

    def __repr__(self):
        return "<Vendor: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')

    def add_contact(self, person):
        contact = VendorContact(person=person)
        self.contacts.append(contact)

    def add_email_address(self, address, type='Info'):
        email = VendorEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Voice'):
        phone = VendorPhoneNumber(number=number, type=type)
        self.phones.append(phone)

Vendor.emails = relationship(
    VendorEmailAddress,
    backref='vendor',
    primaryjoin=VendorEmailAddress.parent_uuid == Vendor.uuid,
    foreign_keys=[VendorEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.email = relationship(
    VendorEmailAddress,
    primaryjoin=and_(
        VendorEmailAddress.parent_uuid == Vendor.uuid,
        VendorEmailAddress.preference == 1),
    foreign_keys=[VendorEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Vendor.phones = relationship(
    VendorPhoneNumber,
    backref='vendor',
    primaryjoin=VendorPhoneNumber.parent_uuid == Vendor.uuid,
    foreign_keys=[VendorPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.phone = relationship(
    VendorPhoneNumber,
    primaryjoin=and_(
        VendorPhoneNumber.parent_uuid == Vendor.uuid,
        VendorPhoneNumber.preference == 1),
    foreign_keys=[VendorPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)

Vendor.contact = relationship(
    VendorContact,
    primaryjoin=and_(
        VendorContact.vendor_uuid == Vendor.uuid,
        VendorContact.preference == 1),
    uselist=False,
    viewonly=True)


class ProductCost(Base):
    """
    Represents a source from which a product may be obtained via purchase.
    """

    __tablename__ = 'product_costs'

    uuid = uuid_column()
    product_uuid = Column(String(32), ForeignKey('products.uuid'), nullable=False)
    vendor_uuid = Column(String(32), ForeignKey('vendors.uuid'), nullable=False)
    preference = Column(Integer, nullable=False)
    code = Column(String(15))

    case_size = Column(Integer)
    case_cost = Column(Numeric(9,5))
    pack_size = Column(Integer)
    pack_cost = Column(Numeric(9,5))
    unit_cost = Column(Numeric(9,5))
    effective = Column(DateTime)

    # case_pack = Column(Integer)
    # case_qty = Column(Integer)
    # pack_qty = Column(Integer)
    # # suggested_retail = Column(Numeric(6,2))
    # case_cost = Column(Numeric(9,5))
    # unit_cost = Column(Numeric(9,5))
    # # cost_effective = Column(DateTime)
    # # cost_expires = Column(DateTime)
    # # cost_recorded = Column(DateTime, default=datetime.datetime.now)

    vendor = relationship(Vendor)

    def __repr__(self):
        return "<ProductCost: %s : %s>" % (self.product, self.vendor)


class ProductPrice(Base):
    """
    Represents a price for a product.
    """

    __tablename__ = 'product_prices'

    uuid = uuid_column()
    product_uuid = Column(String(32), ForeignKey('products.uuid'), nullable=False)
    type = Column(Integer)
    level = Column(Integer)
    starts = Column(DateTime)
    ends = Column(DateTime)
    price = Column(Numeric(8,3))
    multiple = Column(Integer)
    pack_price = Column(Numeric(8,3))
    pack_multiple = Column(Integer)

    def __repr__(self):
        return "<ProductPrice: %s : %s>" % (self.product, self.price)


class Product(Base):
    """
    Represents a product for sale and/or purchase.
    """

    __tablename__ = 'products'

    uuid = uuid_column()
    upc = Column(BigInteger, index=True)
    department_uuid = Column(String(32), ForeignKey('departments.uuid'))
    subdepartment_uuid = Column(String(32), ForeignKey('subdepartments.uuid'))
    category_uuid = Column(String(32), ForeignKey('categories.uuid'))
    brand_uuid = Column(String(32), ForeignKey('brands.uuid'))
    description = Column(String(60))
    description2 = Column(String(60))
    size = Column(String(30))
    unit_of_measure = Column(String(4))

    regular_price_uuid = Column(
        String(32),
        ForeignKey('product_prices.uuid',
                   use_alter=True,
                   name='products_regular_price_uuid_fkey'))

    current_price_uuid = Column(
        String(32),
        ForeignKey('product_prices.uuid',
                   use_alter=True,
                   name='products_current_price_uuid_fkey'))

    # regular_price = Column(Numeric(10,4))
    # # package_price = Column(Numeric(10,4))
    # # package_price_quantity = Column(Integer)
    # sale_price = Column(Numeric(10,4))
    # sale_price_quantity = Column(Integer)

    # case_quantity = Column(Integer)
    # pack_quantity = Column(Integer)
    # pack_price = Column(Numeric(8,3))

    department = relationship(Department)
    subdepartment = relationship(Subdepartment)
    category = relationship(Category)
    brand = relationship(Brand)
    costs = relationship(ProductCost, backref='product',
                         collection_class=ordering_list('preference', count_from=1),
                         order_by=ProductCost.preference)

    def __repr__(self):
        return "<Product: %s>" % self.description

    def __unicode__(self):
        return unicode(self.description or '')

Product.prices = relationship(
    ProductPrice, backref='product',
    primaryjoin=ProductPrice.product_uuid == Product.uuid)

Product.regular_price = relationship(
    ProductPrice,
    primaryjoin=Product.regular_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)

Product.current_price = relationship(
    ProductPrice,
    primaryjoin=Product.current_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)


class EmployeeEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Employee`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Employee'}


class EmployeePhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Employee`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Employee'}


class Employee(Base):
    """
    Represents an employee within the organization.
    """

    __tablename__ = 'employees'

    uuid = uuid_column()
    id = Column(Integer)
    person_uuid = Column(String(32), ForeignKey('people.uuid'), nullable=False)
    status = Column(Integer)
    display_name = Column(String(100))

    person = relationship(Person)

    first_name = association_proxy('person', 'first_name')
    last_name = association_proxy('person', 'last_name')

    def __repr__(self):
        return "<Employee: %s>" % self.person

    def __unicode__(self):
        return unicode(self.display_name or self.person)

    def add_email_address(self, address, type='Home'):
        email = EmployeeEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = EmployeePhoneNumber(number=number, type=type)
        self.phones.append(phone)

Employee.emails = relationship(
    EmployeeEmailAddress,
    backref='employee',
    primaryjoin=EmployeeEmailAddress.parent_uuid == Employee.uuid,
    foreign_keys=[EmployeeEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=EmployeeEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Employee.email = relationship(
    EmployeeEmailAddress,
    primaryjoin=and_(
        EmployeeEmailAddress.parent_uuid == Employee.uuid,
        EmployeeEmailAddress.preference == 1),
    foreign_keys=[EmployeeEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Employee.phones = relationship(
    EmployeePhoneNumber,
    backref='employee',
    primaryjoin=EmployeePhoneNumber.parent_uuid == Employee.uuid,
    foreign_keys=[EmployeePhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=EmployeePhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Employee.phone = relationship(
    EmployeePhoneNumber,
    primaryjoin=and_(
        EmployeePhoneNumber.parent_uuid == Employee.uuid,
        EmployeePhoneNumber.preference == 1),
    foreign_keys=[EmployeePhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


class CustomerPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


class CustomerPerson(Base):
    """
    Represents the association between a :class:`Person` and a customer account
    (:class:`Customer`).
    """

    __tablename__ = 'customers_people'

    uuid = uuid_column()
    customer_uuid = Column(String(32), ForeignKey('customers.uuid'))
    person_uuid = Column(String(32), ForeignKey('people.uuid'))
    ordinal = Column(Integer, nullable=False)

    person = relationship(Person)


class Customer(Base):
    """
    Represents a customer account.  Customer accounts may consist of more than
    one :class:`Person`, in some cases.
    """

    __tablename__ = 'customers'

    uuid = uuid_column()
    id = Column(String(20))
    name = Column(String(255))

    def __repr__(self):
        return "<Customer: %s, %s>" % (self.id, self.name or self.person)

    def __unicode__(self):
        return unicode(self.name or self.person)

    def add_email_address(self, address, type='Home'):
        email = CustomerEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = CustomerPhoneNumber(number=number, type=type)
        self.phones.append(phone)

Customer.emails = relationship(
    CustomerEmailAddress,
    backref='customer',
    primaryjoin=CustomerEmailAddress.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.email = relationship(
    CustomerEmailAddress,
    primaryjoin=and_(
        CustomerEmailAddress.parent_uuid == Customer.uuid,
        CustomerEmailAddress.preference == 1),
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Customer.phones = relationship(
    CustomerPhoneNumber,
    backref='customer',
    primaryjoin=CustomerPhoneNumber.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.phone = relationship(
    CustomerPhoneNumber,
    primaryjoin=and_(
        CustomerPhoneNumber.parent_uuid == Customer.uuid,
        CustomerPhoneNumber.preference == 1),
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)

Customer._people = relationship(
    CustomerPerson, backref='customer',
    primaryjoin=CustomerPerson.customer_uuid == Customer.uuid,
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerPerson.ordinal)

Customer.people = association_proxy(
    '_people', 'person',
    getset_factory=getset_factory,
    creator=lambda x: CustomerPerson(person=x))

Customer._person = relationship(
    CustomerPerson,
    primaryjoin=and_(
        CustomerPerson.customer_uuid == Customer.uuid,
        CustomerPerson.ordinal == 1),
    uselist=False,
    viewonly=True)

Customer.person = association_proxy(
    '_person', 'person',
    getset_factory=getset_factory)


class CustomerGroup(Base):
    """
    Represents an arbitrary group to which :class:`Customer` instances may (or
    may not) belong.
    """

    __tablename__ = 'customer_groups'

    uuid = uuid_column()
    id = Column(String(20))
    name = Column(String(255))

    def __repr__(self):
        return "<CustomerGroup: %s, %s>" % (self.id, self.name)

    def __unicode__(self):
        return unicode(self.name or '')


class CustomerGroupAssignment(Base):
    """
    Represents the assignment of a :class:`Customer` to a
    :class:`CustomerGroup`.
    """

    __tablename__ = 'customers_groups'

    uuid = uuid_column()
    customer_uuid = Column(String(32), ForeignKey('customers.uuid'))
    group_uuid = Column(String(32), ForeignKey('customer_groups.uuid'))
    ordinal = Column(Integer, nullable=False)

    group = relationship(CustomerGroup)

Customer._groups = relationship(
    CustomerGroupAssignment, backref='customer',
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerGroupAssignment.ordinal)

Customer.groups = association_proxy(
    '_groups', 'group',
    getset_factory=getset_factory,
    creator=lambda x: CustomerGroupAssignment(group=x))



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.sil`` -- Standard Interchange Language

Please see the `Standard Interchange Language Specifications
<http://productcatalog.gs1us.org/Store/tabid/86/CategoryID/21/List/1/catpageindex/2/Level/a/ProductID/46/Default.aspx>`_
for more information.
"""

import datetime
from decimal import Decimal

import edbob

import rattail


def val(value):
    """
    Returns a string version of ``value``, suitable for inclusion within a data
    row of a SIL batch.  The conversion is done as follows:

    If ``value`` is ``None``, an empty string is returned.

    If it is an ``int`` or ``decimal.Decimal`` instance, it is converted
    directly to a string (i.e. not quoted).

    If it is a ``datetime.date`` instance, it will be formatted as ``'%Y%j'``.

    If it is a ``datetime.time`` instance, it will be formatted as ``'%H%M'``.

    Otherwise, it is converted to a string if necessary, and quoted with
    apostrophes escaped.
    """

    if value is None:
        return ''
    if isinstance(value, int):
        return str(value)
    if isinstance(value, Decimal):
        return str(value)
    if isinstance(value, datetime.date):
        return value.strftime('%Y%j')
    if isinstance(value, datetime.time):
        return value.strftime('%H%M')
    if not isinstance(value, basestring):
        value = str(value)
    return "'%s'" % value.replace("'", "''")


def consume_batch_id():
    """
    Returns the next available batch identifier, incrementing the number to
    preserve uniqueness.
    """

    config = edbob.AppConfigParser('rattail')
    config_path = config.get_user_file('rattail.conf', create=True)
    config.read(config_path)

    batch_id = config.get('rattail.sil', 'next_batch_id', default='')
    if not batch_id.isdigit():
        batch_id = '1'
    batch_id = int(batch_id)

    config.set('rattail.sil', 'next_batch_id', str(batch_id + 1))
    config_file = open(config_path, 'w')
    config.write(config_file)
    config_file.close()
    return '%08u' % batch_id


def write_batch_header(fileobj, H03='RATAIL', **kwargs):
    """
    Writes a SIL batch header string to ``fileobj``.  All keyword arguments
    correspond to the SIL specification for the Batch Header Dictionary.

    If you do not override ``H03`` (Source Identifier), then Rattail will
    provide a default value for ``H20`` (Software Revision) - that is, unless
    you've supplied it yourself.

    **Batch Header Dictionary:**

    ====  ====    ====  ===========
    Name  Type    Size  Description
    ====  ====    ====  ===========
    H01   CHAR       2  Batch Type
    H02   CHAR       8  Batch Identifier
    H03   CHAR       6  Source Identifier
    H04   CHAR       6  Destination Identifier
    H05   CHAR      12  Audit File Name
    H06   CHAR      12  Response File Name
    H07   DATE       7  Origin Date
    H08   TIME       4  Origin Time
    H09   DATE       7  Execution (Apply) Date
    H10   DATE       4  Execution (Apply) Time
    H11   DATE       7  Purge Date
    H12   CHAR       6  Action Type
    H13   CHAR      50  Batch Description
    H14   CHAR      30  User Defined
    H15   CHAR      30  User Defined
    H16   CHAR      30  User Defined
    H17   NUMBER     1  Warning Level
    H18   NUMBER     5  Maximum Error Count
    H19   CHAR       7  SIL Level/Revision
    H20   CHAR       4  Software Revision
    H21   CHAR      50  Primary Key
    H22   CHAR     512  System Specific Command
    H23   CHAR       8  Dictionary Revision

    Consult the SIL Specification for more information.
    """

    kw = kwargs

    # Provide default for H20 if batch origin is 'RATAIL'.
    H20 = kw.get('H20')
    if H03 == 'RATAIL' and H20 is None:
        H20 = rattail.__version__[:4]

    # Don't quote H09 if special "immediate" value.
    H09 = kw.get('H09')
    if H09 != '0000000':
        H09 = val(H09)

    # Don't quote H10 if special "immediate" value.
    H10 = kw.get('H10')
    if H10 != '0000':
        H10 = val(H10)
    
    row = [
        val(kw.get('H01')),
        val(kw.get('H02')),
        val(H03),
        val(kw.get('H04')),
        val(kw.get('H05')),
        val(kw.get('H06')),
        val(kw.get('H07')),
        val(kw.get('H08')),
        H09,
        H10,
        val(kw.get('H11')),
        val(kw.get('H12')),
        val(kw.get('H13')),
        val(kw.get('H14')),
        val(kw.get('H15')),
        val(kw.get('H16')),
        val(kw.get('H17')),
        val(kw.get('H18')),
        val(kw.get('H19')),
        val(H20),
        val(kw.get('H21')),
        val(kw.get('H22')),
        val(kw.get('H23')),
        ]
        
    fileobj.write('INSERT INTO HEADER_DCT VALUES\n')
    write_row(fileobj, row, quote=False, last=True)
    fileobj.write('\n')


def write_row(fileobj, row, quote=True, last=False):
    """
    Writes a SIL row string to ``fileobj``.

    ``row`` should be a sequence of values.

    If ``quote`` is ``True``, each value in ``row`` will be ran through the
    :func:`val()` function before being written.  If it is ``False``, the
    values are written as-is.

    If ``last`` is ``True``, then ``';'`` will be used as the statement
    terminator; otherwise ``','`` is used.
    """

    terminator = ';' if last else ','
    if quote:
        row = [val(x) for x in row]
    fileobj.write('(' + ','.join(row) + ')' + terminator + '\n')


def write_rows(fileobj, rows):
    """
    Writes a set of SIL row strings to ``fileobj``.

    ``rows`` should be a sequence of sequences, each of which should be
    suitable for use with :func:`write_row()`.

    (This funcion primarily exists to handle the mundane task of setting the
    ``last`` flag when calling :func:`write_row()`.)
    """

    last = len(rows) - 1
    for i, row in enumerate(rows):
        write_row(fileobj, row, last=i == last)


# # from pkg_resources import iter_entry_points

# # import rattail
# # from rattail.batch import make_batch, RattailBatchTerminal
# from rattail.batches import RattailBatchTerminal


# # _junctions = None


# # class SILError(Exception):
# #     """
# #     Base class for SIL errors.
# #     """

# #     pass
    

# # class ElementRequiredError(SILError):
# #     """
# #     Raised when a batch import or export is attempted, but the element list
# #     supplied is missing a required element.
# #     """

# #     def __init__(self, required, using):
# #         self.required = required
# #         self.using = using

# #     def __str__(self):
# #         return "The element list supplied is missing required element '%s': %s" % (
# #             self.required, self.using)


# def default_display(field):
#     """
#     Returns the default UI display value for a SIL field, according to the
#     Rattail field map.
#     """

#     return RattailBatchTerminal.fieldmap_user[field]


# # def get_available_junctions():
# #     """
# #     Returns a dictionary of available :class:`rattail.BatchJunction` classes,
# #     keyed by entry point name.
# #     """

# #     global _junctions
# #     if _junctions is None:
# #         _junctions = {}
# #         for entry_point in iter_entry_points('rattail.batch_junctions'):
# #             _junctions[entry_point.name] = entry_point.load()
# #     return _junctions


# # def get_junction_display(name):
# #     """
# #     Returns the ``display`` value for a registered
# #     :class:`rattail.BatchJunction` class, given its ``name``.
# #     """

# #     juncs = get_available_junctions()
# #     if name in juncs:
# #         return juncs[name].display
# #     return None



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db`` -- Database Stuff
"""

import edbob
# from edbob.db.extensions import activate_extension

import rattail


def init(config):
    """
    Initialize the Rattail database framework.
    """

    from rattail.db.extension import model
    edbob.graft(rattail, model)

    from rattail.db.extension import enum
    edbob.graft(rattail, enum)


# def init_database(engine, session):
#     """
#     Initialize an ``edbob`` database for use with Rattail.
#     """

#     activate_extension('rattail', engine)

#     columns = [
#         ('F01',         'UPC',                  'GPC(14)'),
#         ('F02',         'Description',          'CHAR(20)'),
#         ('F03',         'Department Number',    'NUMBER(4,0)'),
#         ('F22',         'Size',                 'CHAR(30)'),
#         ('F155',        'Brand',                'CHAR(30)'),
#         ('F238',        'Department Name',      'CHAR(30)'),
#         ]

#     for name, disp, dtype in columns:
#         session.add(rattail.SilColumn(
#                 sil_name=name, display=disp, data_type=dtype))
#         session.flush()

#     dictionaries = [
#         ('DEPT_DCT', 'Department', [
#                 ('F03', True),
#                 'F238',
#                 ]),
#         ('ITEM_DCT', 'Product', [
#                 ('F01', True),
#                 'F02',
#                 'F03',
#                 'F22',
#                 'F155',
#                 ]),
#         # ('PRICE_DCT', 'Price', []),
#         # ('FCOST_DCT', 'Future Cost', []),
#         # ('FSPRICE_DCT', 'Future Sale Price', []),
#         # ('CLASS_GROUP', 'Scale Class / Group', []),
#         # ('NUTRITION', 'Scale Nutrition', []),
#         # ('SCALE_TEXT', 'Scale Text', []),
#         # ('VENDOR_DCT', 'Vendor', []),
#         ]

#     for name, desc, cols in dictionaries:
#         bd = rattail.BatchDictionary(name=name, description=desc)
#         for col in cols:
#             key = False
#             if not isinstance(col, basestring):
#                 col, key = col
#             q = session.query(rattail.SilColumn)
#             q = q.filter(rattail.SilColumn.sil_name == col)
#             col = q.one()
#             bd.columns.append(
#                 rattail.BatchDictionaryColumn(sil_column=col, key=key))
#         session.add(bd)
#         session.flush()



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.extension.enum`` -- Enumerations
"""


BATCH_ADD                       = 'ADD'
BATCH_ADD_REPLACE               = 'ADDRPL'
BATCH_CHANGE                    = 'CHANGE'
BATCH_LOAD                      = 'LOAD'
BATCH_REMOVE                    = 'REMOVE'

BATCH_ACTION_TYPE = {
    BATCH_ADD                   : "Add",
    BATCH_ADD_REPLACE           : "Add/Replace",
    BATCH_CHANGE                : "Change",
    BATCH_LOAD                  : "Load",
    BATCH_REMOVE                : "Remove",
    }


# BATCH_MAIN_ITEM                 = 'ITEM_DCT'

# BATCH_DICTIONARY = {
#     BATCH_MAIN_ITEM             : "Main Item",
#     }


EMPLOYEE_STATUS_CURRENT         = 1
EMPLOYEE_STATUS_FORMER          = 2

EMPLOYEE_STATUS = {
    EMPLOYEE_STATUS_CURRENT     : "current",
    EMPLOYEE_STATUS_FORMER      : "former",
    }


PRICE_TYPE_REGULAR              = 0
PRICE_TYPE_TPR                  = 1
PRICE_TYPE_SALE                 = 2
PRICE_TYPE_MANAGER_SPECIAL      = 3
PRICE_TYPE_ALTERNATE            = 4
PRICE_TYPE_FREQUENT_SHOPPER     = 5
PRICE_TYPE_MFR_SUGGESTED        = 901

PRICE_TYPE = {
    PRICE_TYPE_REGULAR          : "Regular Price",
    PRICE_TYPE_TPR              : "TPR",
    PRICE_TYPE_SALE             : "Sale",
    PRICE_TYPE_MANAGER_SPECIAL  : "Manager Special",
    PRICE_TYPE_ALTERNATE        : "Alternate Price",
    PRICE_TYPE_FREQUENT_SHOPPER : "Frequent Shopper",
    PRICE_TYPE_MFR_SUGGESTED    : "Manufacturer's Suggested",
    }


UNIT_OF_MEASURE_EACH            = '01'
UNIT_OF_MEASURE_POUND           = '49'

UNIT_OF_MEASURE = {
    UNIT_OF_MEASURE_EACH        : "Each",
    UNIT_OF_MEASURE_POUND       : "Pound",
    }


# VENDOR_CATALOG_NOT_PARSED       = 1
# VENDOR_CATALOG_PARSED           = 2
# VENDOR_CATALOG_COGNIZED         = 3
# VENDOR_CATALOG_PROCESSED        = 4

# VENDOR_CATALOG_STATUS = {
#     VENDOR_CATALOG_NOT_PARSED   : "not parsed",
#     VENDOR_CATALOG_PARSED       : "parsed",
#     VENDOR_CATALOG_COGNIZED     : "cognized",
#     VENDOR_CATALOG_PROCESSED    : "processed",
#     }



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.labels`` -- Label Printing
"""

import os
import os.path
import socket
from cStringIO import StringIO

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import edbob
from edbob.util import requires_impl


_profiles = OrderedDict()


class LabelPrinter(edbob.Object):
    """
    Base class for all label printers.

    Label printing devices which are "natively" supported by Rattail will each
    derive from this class in order to provide implementation details specific
    to the device.  You will typically instantiate one of those subclasses (or
    one of your own design) in order to send labels to your physical printer.
    """

    profile_name = None
    formatter = None

    @requires_impl()
    def print_labels(self, labels, *args, **kwargs):
        """
        Prints labels found in ``labels``.
        """

        pass


class CommandFilePrinter(LabelPrinter):
    """
    Generic :class:`LabelPrinter` subclass which "prints" labels to a file in
    the form of native printer (textual) commands.  The output file is then
    expected to be picked up by a file monitor, and finally sent to the printer
    from there.
    """

    def batch_header_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will be the first
        which are written to the file.
        """

        return None

    def batch_footer_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will be the last
        which are written to the file.
        """

        return None

    def print_labels(self, labels, output_dir=None):
        """
        "Prints" ``labels`` by generating a command file in the output folder.
        The full path of the output file to which commands are written will be
        returned to the caller.

        If ``output_dir`` is not specified, and the printer instance is
        associated with a :class:`LabelProfile` instance, then config will be
        consulted for the output path.  If a path still is not found, the
        current (working) directory will be assumed.
        """

        if not output_dir and self.profile_name:
            output_dir = edbob.config.get('rattail.labels', '%s.output_dir' % self.profile_name)
        if not output_dir:
            output_dir = os.getcwd()

        fn = '%s_%s.labels' % (socket.gethostname(),
                               edbob.local_time().strftime('%Y-%m-%d_%H-%M-%S'))
        labels_path = os.path.join(output_dir, fn)
        labels_file = open(labels_path, 'w')

        header = self.batch_header_commands()
        if header:
            print >> labels_file, '\n'.join(header)

        print >> labels_file, self.formatter.format_labels(labels)

        footer = self.batch_footer_commands()
        if footer:
            print >> labels_file, '\n'.join(footer)

        labels_file.close()
        return labels_path


class LabelFormatter(edbob.Object):
    """
    Base class for all label formatters.
    """

    format = None

    @requires_impl()
    def format_labels(self, labels, *args, **kwargs):
        """
        Formats ``labels`` and returns the result.
        """

        pass


class CommandFormatter(LabelFormatter):
    """
    Generic subclass of :class:`LabelFormatter` which generates native printer
    (textual) commands.
    """

    def label_header_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will immediately
        precede each *label* in one-up printing, and immediately precede each
        *label pair* in two-up printing.
        """

        return None

    @requires_impl()
    def label_body_commands(self):
        pass

    def label_footer_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will immedately
        follow each *label* in one-up printing, and immediately follow each
        *label pair* in two-up printing.
        """

        return None


class TwoUpCommandFormatter(CommandFormatter):
    """
    Generic subclass of :class:`LabelFormatter` which generates native printer
    (textual) commands.

    This class contains logic to implement "two-up" label printing.
    """

    @property
    @requires_impl(is_property=True)
    def half_offset(self):
        """
        The X-coordinate value by which the second label should be offset, when
        two labels are printed side-by-side.
        """

        pass

    def format_labels(self, labels):
        fmt = StringIO()

        half_started = False
        for product, quantity in labels:
            for i in range(quantity):
                if half_started:
                    print >> fmt, '\n'.join(self.label_body_commands(product, x=self.half_offset))
                    footer = self.label_footer_commands()
                    if footer:
                        print >> fmt, '\n'.join(footer)
                    half_started = False
                else:
                    header = self.label_header_commands()
                    if header:
                        print >> fmt, '\n'.join(header)
                    print >> fmt, '\n'.join(self.label_body_commands(product, x=0))
                    half_started = True

        if half_started:
            footer = self.label_footer_commands()
            if footer:
                print >> fmt, '\n'.join(footer)

        val = fmt.getvalue()
        fmt.close()
        return val
    

class LabelProfile(edbob.Object):
    """
    Represents a label printing profile.  This abstraction is used to define
    not only the physical (or otherwise?) device to which label should be sent,
    but the label formatting specifics as well.
    """

    name = None
    display_name = None
    printer_factory = None
    formatter_factory = None
    format = None

    def get_formatter(self):
        if self.formatter_factory:
            return self.formatter_factory(format=self.format)
        return None

    def get_printer(self):
        if self.printer_factory:
            return self.printer_factory(
                profile_name=self.name,
                formatter=self.get_formatter())
        return None


def init(config):
    """
    Initializes the label printing system.

    This reads label profiles from config and caches the corresponding
    :class:`LabelProfile` instances in memory.
    """

    profiles = config.require('rattail.labels', 'profiles')
    profiles = profiles.split(',')
    for key in profiles:
        key = key.strip()
        profile = LabelProfile(name=key)
        profile.display_name = config.require('rattail.labels', '%s.display' % key)
        profile.printer_factory = edbob.load_spec(
            config.require('rattail.labels', '%s.printer' % key))
        profile.formatter_factory = edbob.load_spec(
            config.require('rattail.labels', '%s.formatter' % key))
        profile.format = config.get('rattail.labels', '%s.format' % key)
        _profiles[key] = profile


def get_profile(name):
    """
    Returns the :class:`LabelProfile` instance corresponding to ``name``.
    """

    return _profiles.get(name)


def iter_profiles():
    """
    Returns an iterator over the collection of :class:`LabelProfile` instances
    which were read and created from config.
    """

    return _profiles.itervalues()



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.batches`` -- Batch Interface
"""

# import logging

from sqlalchemy import and_
from sqlalchemy.orm import object_session

import edbob
from edbob.db import needs_session
from edbob.util import requires_impl

import rattail
# from rattail.db import needs_session
# from rattail.util import get_entry_points, requires_impl


# log = logging.getLogger(__name__)
# # _registered_sources = None


# # def get_registered_sources():
# #     """
# #     Returns the entry point map for registered batch sources.
# #     """

# #     global _registered_sources
# #     if _registered_sources is None:
# #         _registered_sources = get_entry_points('rattail.batch_sources')
# #     return _registered_sources


# # @needs_session
# # def get_sources(session):
# #     """
# #     Returns a dictionary of registered :class:`rattail.BatchSource` classes,
# #     keyed by :attr:`BatchSource.name`.
# #     """

# #     sources = {}
# #     for src in session.query(rattail.BatchSource):
# #         sources[src.name] = src
# #     return sources


class BatchTerminal(edbob.Object):
    """
    Defines the interface for data batch terminals.  Subclass this when
    implementing new data awareness and/or integration with external systems.
    """

    @property
    @requires_impl()
    def name(self):
        pass

    @property
    @requires_impl()
    def display(self):
        pass

    @property
    @requires_impl()
    def fieldmap_internal(self):
        pass

    @property
    @requires_impl()
    def fieldmap_user(self):
        pass

    @requires_impl()
    def provide_rows(self, session, rowclass, dictionary, query=None, **kwargs):
        """
        Generator which yields (new) batch row instances.  This is used to
        populate batches.
        """

        raise NotImplementedError

    # def get_elements(self, elements, fieldmap, *required):
    #     """
    #     Returns the proper element list according to current context.
    #     """

    #     if elements is None:
    #         elements = sorted(fieldmap)
    #     self.require_elements(elements, *required)
    #     return elements

    # def require_elements(self, using, *required):
    #     """
    #     Officially require one or more elements when processing a batch.
    #     """

    #     for elements in required:
    #         elements = elements.split(',')
    #         for element in elements:    
    #             if element not in using:
    #                 raise sil.ElementRequiredError(element, using)


class RattailBatchTerminal(BatchTerminal):
    """
    Defines the core batch terminal for Rattail.
    """
    
    name = 'rattail'
    description = "Rattail (local)"

    source_columns = {
        'ITEM_DCT': [
            'F01',
            'F02',
            ],
        }

    target_columns = {
        'DEPT_DCT': [
            'F03',
            'F238',
            ],
        'ITEM_DCT': [
            'F01',
            'F02',
            'F03',
            'F22',
            'F155',
            ],
        }

    def provide_rows(self, session, rowclass, dictionary, query=None, **kwargs):

        if dictionary.name == 'DEPT_DCT':
            if not query:
                query = session.query(rattail.Department)
            for dept in query:
                yield rowclass(
                    F03=dept.number,
                    F238=dept.name,
                    )
            return

        elif dictionary.name == 'ITEM_DCT':
            if not query:
                query = session.query(rattail.Product)
            for product in query:
                yield rowclass(
                    F01=int(product.upc),
                    F02=product.description[:20],
                    F155=product.brand.name if product.brand else None,
                    )
            return

        assert False, "FIXME"

    # def import_main_item(self, session, elements=None, progress_factory=None):
    #     """
    #     Create a main item (ITEM_DCT) batch from current Rattail data.
    #     """

    #     elements = self.get_elements(elements, self.fieldname_map_main_item, 'F01')
    #     batch = make_batch(self.name, elements, session,
    #                        description="Main item data from Rattail")

    #     products = session.query(rattail.Product)
    #     prog = None
    #     if progress_factory:
    #         prog = progress_factory("Creating batch", products.count())
    #     for i, prod in enumerate(products, 1):
    #         fields = {}
    #         for name in elements:
    #             fields[name] = row[self.fieldname_map_main_item[name]]
    #         batch.append(**fields)
    #         if prog:
    #             prog.update(i)
    #     if prog:
    #         prog.destroy()

    #     return batch

    def add_departments(self, session, batch):
        for row in batch.provide_rows():
            dept = rattail.Department()
            dept.number = row.F03
            dept.name = row.F238
            session.add(dept)
            session.flush()

    def add_replace_departments(self, session, batch):
        for row in batch.provide_rows():
            q = session.query(rattail.Department)
            q = q.filter_by(number=row.F03)
            if q.count():
                prods = session.query(rattail.Product)
                prods = prods.filter_by(department_uuid=q.first().uuid)
                if prods.count():
                    prods.update(dict(department_uuid=None), synchronize_session='fetch')
                q.delete(synchronize_session=False)

            dept = rattail.Department()
            dept.number = row.F03
            dept.name = row.F238
            session.add(dept)
            session.flush()

    def add_products(self, session, batch):
        q = session.query(rattail.Department)
        depts = {}
        for dept in q:
            depts[dept.number] = dept

        q = session.query(rattail.Brand)
        brands = {}
        for brand in q:
            brands[brand.name] = brand

        for row in batch.provide_rows():
            dept = None
            if row.F03:
                if row.F03 in depts:
                    dept = depts[row.F03]
                else:
                    dept = rattail.Department(number=row.F03)
                    session.add(dept)
                    session.flush()
                    depts[dept.number] = dept

            brand = None
            if row.F155:
                if row.F155 in brands:
                    brand = brands[row.F155]
                else:
                    brand = rattail.Brand(name=row.F155)
                    session.add(brand)
                    session.flush()
                    brands[brand.name] = brand

            prod = rattail.Product()
            prod.upc = row.F01
            prod.description = row.F02
            prod.size = row.F22
            prod.department = dept
            prod.brand = brand
            session.add(prod)
            session.flush()

    def add_replace_products(self, session, batch):
        q = session.query(rattail.Department)
        depts = {}
        for dept in q:
            depts[dept.number] = dept

        q = session.query(rattail.Brand)
        brands = {}
        for brand in q:
            brands[brand.name] = brand

        products = session.query(rattail.Product)
        for row in batch.provide_rows():
            dept = None
            if row.F03:
                if row.F03 in depts:
                    dept = depts[row.F03]
                else:
                    dept = rattail.Department(number=row.F03)
                    session.add(dept)
                    session.flush()
                    depts[dept.number] = dept

            brand = None
            if row.F155:
                if row.F155 in brands:
                    brand = brands[row.F155]
                else:
                    brand = rattail.Brand(name=row.F155)
                    session.add(brand)
                    session.flush()
                    brands[brand.name] = brand

            q = products.filter_by(upc=row.F01)
            if q.count():
                q.delete(synchronize_session=False)

            prod = rattail.Product()
            prod.upc = row.F01
            prod.description = row.F02
            prod.size = row.F22
            prod.department = dept
            prod.brand = brand
            session.add(prod)
            session.flush()

    def change_departments(self, session, batch):
        depts = session.query(rattail.Department)
        for row in batch.provide_rows():
            dept = depts.filter_by(number=row.F03).first()
            if dept:
                dept.name = row.F238
                session.flush()

    def change_products(self, session, batch):
        q = session.query(rattail.Department)
        depts = {}
        for dept in q:
            depts[dept.number] = dept

        products = session.query(rattail.Product)
        for row in batch.provide_rows():
            prod = products.filter_by(upc=row.F01).first()
            if prod:
                dept = None
                if row.F03:
                    if row.F03 in depts:
                        dept = depts[row.F03]
                    else:
                        dept = rattail.Department(number=row.F03)
                        session.add(dept)
                        session.flush()
                        depts[dept.number] = dept
                prod.dept = dept
                prod.size = row.F22
                session.flush()

    def execute_batch(self, batch):
        """
        Executes ``batch``, which should be a :class:`rattail.Batch` instance.
        """

        session = object_session(batch)

        if batch.action_type == rattail.BATCH_ADD:

            if batch.dictionary.name == 'DEPT_DCT':
                self.add_departments(session, batch)
                return
            if batch.dictionary.name == 'ITEM_DCT':
                self.add_products(session, batch)
                return

        if batch.action_type == rattail.BATCH_ADD_REPLACE:
            if batch.dictionary.name == 'DEPT_DCT':
                return self.add_replace_departments(session, batch)
            if batch.dictionary.name == 'ITEM_DCT':
                return self.add_replace_products(session, batch)

        if batch.action_type == rattail.BATCH_CHANGE:

            if batch.dictionary.name == 'DEPT_DCT':
                self.change_departments(session, batch)
                return

            if batch.dictionary.name == 'ITEM_DCT':
                return self.change_products(session, batch)

        if batch.action_type == rattail.BATCH_REMOVE:
            if batch.dictionary.name == 'DEPT_DCT':
                return self.remove_departments(session, batch)
            if batch.dictionary.name == 'ITEM_DCT':
                return self.remove_products(session, batch)

        assert False, "FIXME"

    def remove_departments(self, session, batch):
        depts = session.query(rattail.Department)
        products = session.query(rattail.Product)
        for row in batch.provide_rows():
            dept = depts.filter_by(number=row.F03).first()
            if dept:
                q = products.filter_by(department_uuid=dept.uuid)
                if q.count():
                    q.update({'department_uuid': None}, synchronize_session=False)
                session.delete(dept)
                session.flush()

    def remove_products(self, session, batch):
        products = session.query(rattail.Product)
        for row in batch.provide_rows():
            prod = products.filter_by(upc=row.F01).first()
            if prod:
                session.delete(prod)
                session.flush()


# def make_batch(source, elements, session, batch_id=None, **kwargs):
#     """
#     Create and return a new SIL-based :class:`rattail.Batch` instance.
#     """

#     if not batch_id:
#         batch_id = next_batch_id(source, consume=True)

#     kwargs['source'] = source
#     kwargs['batch_id'] = batch_id
#     kwargs['target'] = target
#     kwargs['elements'] = elements
#     kwargs.setdefault('sil_type', 'HM')
#     kwargs.setdefault('action_type', rattail.BATCH_ADD_REPLACE)
#     kwargs.setdefault('dictionary', rattail.BATCH_MAIN_ITEM)
#     batch = rattail.Batch(**kwargs)

#     session.add(batch)
#     session.flush()
#     batch.table.create()
#     log.info("Created batch table: %s" % batch.table.name)
#     return batch


@needs_session
def next_batch_id(session, source, consume=False):
    """
    Returns the next available batch ID (as an integer) for the given
    ``source`` SIL ID.

    If ``consume`` is ``True``, the "running" ID will be incremented so that
    the next caller will receive a different ID.
    """

    batch_id = edbob.get_setting('batch.next_id.%s' % source,
                                 session=session)
    if batch_id is None or not batch_id.isdigit():
        batch_id = 1
    else:
        batch_id = int(batch_id)

    while True:
        q = session.query(rattail.Batch)
        q = q.join((rattail.BatchTerminal,
                    rattail.BatchTerminal.uuid == rattail.Batch.source_uuid))
        q = q.filter(and_(
                rattail.BatchTerminal.sil_id == source,
                rattail.Batch.batch_id == '%08u' % batch_id,
                ))
        if not q.count():
            break
        batch_id += 1

    if consume:
        edbob.save_setting('batch.next_id.%s' % source, str(batch_id + 1),
                           session=session)
    return batch_id
